##############################################################################
#
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
#
# Copyright (c) Digital Creations.  All rights reserved.
#
# This license has been certified as Open Source(tm).
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
#
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
#
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
#
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
#
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
#
#
# Disclaimer
#
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
#
#
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
#
##############################################################################
from string import lower, split
from Globals import DTMLFile, MessageDialog, InitializeClass
from AccessControl import ClassSecurityInfo
from Acquisition import aq_base, aq_inner, aq_parent
from OFS.PropertyManager import PropertyManager
from ZPublisher import Converters

from Products.CMFCore.WorkflowCore import WorkflowMethod
from Products.CMFCore.utils import getToolByName

import utilities

# Import permission names
import PropertyCorePermissions

def _getNewId(prefix=None):
    """ Obtain new id """
    id = utilities.getNewId()
    if prefix != None:
        id = lower(str(prefix)) + '_' + id
    return id

def InitializeClassLinkPropertyManager(klass
                              , modify_priviledge
                              , property_extensible_schema__=None
                              , property_link_id_schema__=None
                              ):
    """
    Initialize PropertyManager interface and Class with appropriate class security.
    """
    # Declarative security
    if not hasattr(klass, 'security'):
        klass.security = ClassSecurityInfo()

    klass.security.declareProtected(modify_priviledge, 'manage_addProperty')
    klass.security.declareProtected(modify_priviledge, 'manage_delProperties')
    klass.security.declareProtected(modify_priviledge, 'manage_changeProperties')
    klass.security.declareProtected(modify_priviledge, 'manage_editProperties')
    klass.security.declareProtected(modify_priviledge, 'manage_changePropertyTypes')
    klass.security.declareProtected(modify_priviledge, 'manage_firstProperties')
    klass.security.declareProtected(modify_priviledge, 'manage_lastProperties')

    if property_extensible_schema__ != None:
        klass.property_extensible_schema__=property_extensible_schema__
    if property_link_id_schema__ != None:
        klass.property_link_id_schema__=property_link_id_schema__

class LinkPropertyManager(PropertyManager):
    """
    """
    meta_type='CMF Link Property Manager'

    manage_propertiesForm=DTMLFile('dtml/properties', globals(),
                                   property_extensible_schema__=1)
    manage_propertyTypeForm=DTMLFile('dtml/propertyType', globals())

    # Declarative security
    security = ClassSecurityInfo()
    security.declareObjectProtected(PropertyCorePermissions.View)

    #######################################
    # init
    def __init__(self, id, title=None, parent=None
                 , property_extensible_schema__=None
                 , property_link_id_schema__=None):
        self.id = id
        self._properties = PropertyManager._properties
        self.manage_changeProperties(title=title)
        if parent is None:
            self.parents=[(None,self.getId())]
        else:
            self.parents=[(None,parent)]
        if property_extensible_schema__ is not None:
            self.property_extensible_schema__ = property_extensible_schema__
        if property_link_id_schema__ is not None:
            self.property_link_id_schema__ = property_link_id_schema__

    security.declarePublic('getNewId')
    def getNewId(self, prefix=None):
        return _getNewId(prefix)

    security.declarePublic('getParents')
    def getParents(self, link_edit=None, REQUEST=None,RESPONSE=None):
        """
        """
        parents = ''
        for p in self.parents:
            (subject, uri) = p
            parent = self._my_traverse(uri)
            if parent is not None:
                uri = parent._my_absolute_url(link_edit=link_edit)
                subject = self._getPropertyLinkText(subject, parent)[1]
            parents = parents + '<a href="%s"><font color=red>%s</font></a>&nbsp;' % (uri, subject)
        return parents

    #######################################
    # PropertyManager Methods
    def getProperty(self, id, d=None, link_edit=None):
        """Get the property 'id', returning the optional second
           argument or None if no such property is found."""
        if self.hasProperty(id):
            if (self.getPropertyType(id) != 'link' or
                (self.getPropertyType(id) == 'link' and link_edit is None)):
                return getattr(self, id)
            else:
                return self._getPropertyLink(id,link_edit)
        return d

    def manage_firstProperties(self, ids=None, REQUEST=None):
        """Re-sort properties, moving selected ones first"""
        if not getattr(self, 'property_extensible_schema__', None):
            raise 'Unauthorized'

        if ids is None:
            return MessageDialog(
                title='No property specified',
                message='No properties were specified!',
                action ='./manage_propertiesForm',)

        l = self._properties or []
        l = filter(lambda x,ids=ids: x['id'] in ids, l) + \
            filter(lambda x,ids=ids: x['id'] not in ids, l)
        self._properties = l

        if REQUEST is not None:
            return self.manage_propertiesForm(self, REQUEST)

    def manage_lastProperties(self, ids=None, REQUEST=None):
        """Re-sort properties, moving selected ones last"""
        if not getattr(self, 'property_extensible_schema__', None):
            raise 'Unauthorized'

        if ids is None:
            return MessageDialog(
                title='No property specified',
                message='No properties were specified!',
                action ='./manage_propertiesForm',)

        l = self._properties or []
        l = filter(lambda x,ids=ids: x['id'] not in ids, l) + \
            filter(lambda x,ids=ids: x['id'] in ids, l)
        self._properties = l

        if REQUEST is not None:
            return self.manage_propertiesForm(self, REQUEST)

    manage_addProperty         = WorkflowMethod(PropertyManager.manage_addProperty)
    manage_delProperties       = WorkflowMethod(PropertyManager.manage_delProperties)
    manage_changeProperties    = WorkflowMethod(PropertyManager.manage_changeProperties)
    manage_changePropertyTypes = WorkflowMethod(PropertyManager.manage_changePropertyTypes)
    manage_firstProperties     = WorkflowMethod(manage_firstProperties)
    manage_lastProperties      = WorkflowMethod(manage_lastProperties)

    #######################################
    # override subset of converters for PropertyManager
    def _setPropValue(self, id, value):
        if self.getPropertyType(id) != 'link':
            return PropertyManager._setPropValue(self, id, value)
        value = value or ''
        if hasattr(self,id):
            (type_name,subject,uri) = getattr(self,id)
            # modify link - subject
            link = split(value,':',2)
            if len(link) == 1:
                # only modify Subject
                subject = link[0] or None
            elif len(link) == 2:
                # modify Type and Subject, however **RESET** link
                type_name = link[0] or None
                subject = link[1] or None
                uri = None
            elif len(link) == 3:
                # modify Type, Subject, and URI, however **CREATE** new link if necessary
                type_name = link[0] or None
                subject = link[1] or None
                uri = link[2] or None
                if type_name is not None and uri is None:
                    if getattr(self._my_folder(),'property_link_id_schema__',None):
                        uri = 'link_' + id
                    else:
                        uri = self.getNewId(type_name)
                    self._my_folder().invokeFactory(type_name=type_name
                                                    , id=uri
                                                    , parent=self.getId()
                                                    )
            else:
                raise ValueError, 'invalid link (%s)' % (value)
        else:
            # add link - type_name:subject:uri
            link = split(value,':',2)
            if len(link) == 1:
                type_name = link[0]
                subject = None
                uri = None
            elif len(link) == 2:
                type_name = link[0]
                subject = link[1] or None
                uri = None
            elif len(link) == 3:
                type_name = link[0]
                subject = link[1] or None
                if getattr(self._my_folder(),'property_link_id_schema__',None):
                    uri = 'link_' + id
                else:
                    uri = link[2] or self.getNewId(type_name)
                self._my_folder().invokeFactory(type_name=type_name
                                                , id=uri
                                                , parent=self.getId()
                                                )
            else:
                raise ValueError, 'invalid link (%s)' % (value)
        # set Property
        return PropertyManager._setPropValue(self, id, (type_name,subject,uri))

    def field2float(v):
        if not v:
            return None
        else:
            return Converters.field2float(v)
    def field2int(v):
        if not v:
            return None
        else:
            return Converters.field2int(v)
    def field2long(v):
        if not v:
            return None
        else:
            return Converters.field2long(v)
    def field2date(v):
        if not v:
            return None
        else:
            return Converters.field2date(v)
    def field2link(v):
        if hasattr(v,'read'): v=v.read()
        else: v=str(v)
        return v

    type_converters = Converters.type_converters
    type_converters.update({'float'  : field2float
                            , 'int'  : field2int
                            , 'long' : field2long
                            , 'date' : field2date
                            , 'link' : field2link
                            })

    #######################################
    # private
    def _getPropertyLink(self,id,link_edit=0):
        link = ''
        val = getattr(self,id)

        #######################################
        # implies unknown link
        if val is None:
            link = '<font color=red>?</font>'
            return link

        #######################################
        (type_name,subject,uri) = getattr(self,id)
        type = self.portal_types.getTypeInfo(type_name)

        membertool= getToolByName(self, 'portal_membership')
        auname = membertool.getAuthenticatedMember().getUserName()
        ouname = self.getOwner().getUserName()

        #######################################
        # implies construct new object
        if uri is None:
            if (auname == ouname and type in self.allowedContentTypes()):
                link = '<a href="%s/pcore_edit_properties?%s=%s:%s:%s"><font color=red>?</font></a>' % (
                    self.absolute_url(),
                    id,
                    type_name or '', subject or '', '')
            else:
                link = '<font color=red>?</font>'

            return link

        #######################################
        obj = self._my_traverse(uri,None)

        #######################################
        # implies stale (or deleted) link
        if obj is None:
            link = ''
            # reset and create
            if (auname == ouname and type in self.allowedContentTypes()):
                link = link + '<a href="%s/pcore_edit_properties?%s=%s:%s:%s"><font color=red>?</font></a>' % (
                    self.absolute_url(),
                    id,
                    type_name or '', subject or '', '')
            else:
                link = link + '<font color=red>?</font>'
            # reset only
            link = link + '&nbsp;'
            if (auname == ouname and type in self.allowedContentTypes()):
                link = link + '<a href="%s/pcore_edit_properties?%s=%s:%s"><font color=red>X</font></a>' % (
                    self.absolute_url(),
                    id,
                    type_name or '', subject or '')
            else:
                link = link + '<font color=red>X</font>'

            return link

        #######################################
        if link_edit and type is not None:
            subjects = self.portal_metadata.listAllowedSubjects(type)
            # add if existing value is not in allowed subjects
            if subjects is not None:
                if subject is not None and subject not in subjects:
                    subjects = ['subject'] + subjects
            elif subject is not None:
                subjects = ['subject']
            # show only if choice is available
            if subjects is not None and len(subjects):
                # select tag
                link = '<select name="%s:string">' % id
                # null option
                link = link + '<option value=""'
                if subject is None:
                    link = link + ' selected'
                link = link + '></option>\n'
                # allowedSubjects
                for s in subjects:
                    link = link + '<option value="%s"' % s
                    if s == subject:
                        link = link + ' selected'
                    link = link + '>%s</option>\n' % s
                # /select tag
                link = link + '</select> &nbsp;'

        (uri, text) = self._getPropertyLinkText(uri,None,link_edit)
        link = link + '<i>%s</i>&nbsp;<a href="%s"><font color=red>[&nbsp;%s&nbsp;]</font></a>' % ((subject or 'Unfiled'), uri, text)

        return link

    def _getPropertyLinkText(self,uri,obj=None,link_edit=0):
        """ Return appropriate text for link """
        text = uri
        obj = self._my_traverse(uri,obj)
        # get search text
        if obj is not None:
            uri = obj._my_absolute_url(link_edit=link_edit)
            try:
                text = (getattr(obj, 'title', None) or getattr(obj, 'description', None) or
                        obj.SearchableText() or text)
                return (uri,text)
            except:
                raise
                return (uri,text)
        else:
            return (uri,text)

    def _my_folder(self):
        """ Obtain parent folder """
        if self.isPrincipiaFolderish:
            return self
        else:
            return aq_parent(aq_inner(self))

    def _my_traverse(self,uri,obj=None):
        if obj is None:
            # stolen from CMFZWiki
            folder = self._my_folder()
            # if something of this name exists, link to it;
            # otherwise, provide a "?" creation link
            if hasattr(aq_base(folder), uri):
                obj = folder[uri]
            elif hasattr(folder, uri):
                obj = getattr(folder, uri)
            else:
                obj = self.restrictedTraverse(uri, None)

        if obj is not None and not hasattr(obj, 'absolute_url'):
            # Whoops!  Probly an acquired method, which we do *not* want
            # want to link, and which we do *not* want to allow being
            # shadowed by a wiki page.
            obj = None

        return obj

    def _my_absolute_url(self, relative=None, link_edit=None):
        url = self.absolute_url(relative)
        try:
            if link_edit:
                url = '%s/%s' % (url, self.getTypeInfo().getActionById('edit', ''))
            else:
                url = '%s/%s' % (url, self.getTypeInfo().getActionById('view', ''))
        except:
            nop = 1
        return url

    def _extend_properties(self,properties):
        """ Safely extend properties map """
        if hasattr(self,'_properties'):
            old_properties = self._properties
        else:
            old_properties = []

        properties_ids = map(lambda p: p['id'], properties)
        return filter(lambda p,ids=properties_ids: p['id'] not in ids, old_properties) \
               + properties

PropertyManager.type_converters = LinkPropertyManager.type_converters

# Initialize the BaseContent class, setting up security.
InitializeClassLinkPropertyManager(LinkPropertyManager
                                   , PropertyCorePermissions.Modify
                                   , 1
                                   , 0
                                   )
InitializeClass(LinkPropertyManager)
